#!/usr/bin/python
"""    Space Parcels -- ingame routines.
    Copyright (C) 2005 Mathieu "Zig" Brethes <just.want.to.be@free.fr>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

# Main game file. Contains classes, algorithms, and stuff!

import pygame, os.path, math, random, copy
from pygame.locals import *
from sound import SoundEngine
from menu import MainMenu
import myglobals

# the base animator. Pos should be a Rect, sequence a list of at least 1 element.
# angle is not compulsory.
class Animator(pygame.sprite.Sprite):
	def __init__(self, sequence, pos, angle=0):
		pygame.sprite.Sprite.__init__(self)
		self.angle = angle
		self.sequence = sequence
		self.rect = pos
		self.seqIndex = 0

		if angle != 0:
			self.image = pygame.transform.rotate(sequence[self.seqIndex], angle)
		else:
			self.image = sequence[self.seqIndex]

	def update(self):
		self.seqIndex += 1
		if self.seqIndex >= len(self.sequence):
			self.seqIndex = 0
		if self.angle != 0:
			self.image = pygame.transform.rotate(self.sequence[self.seqIndex], self.angle)
		else:
			self.image = self.sequence[self.seqIndex]

	# useful for sprite with multiple sequences.
	def setSequence(self, sequence):
		self.sequence = sequence
		self.seqIndex = 0

# The base gravitator. Pos should be a Rect, sequence a sequence of Surfaces.
# Angle is not compulsory and can be left out.		Animator.update(self)
# Velocity should be a 2-uple of integers, mass an integer.
# isAffected determines whether the gravitator will be attracted by other objects in
# the gamefield. Default is true.
# REFRATE is the refresh rate of the application.
class Gravitator(Animator):
	REFRATE = 30
	
	def __init__(self, sequence, pos, mass, velocity, isAffected=True,angle=0):
		Animator.__init__(self, sequence, pos, angle)
		self.mass = mass
		self.velocityX = velocity[0] / Gravitator.REFRATE
		self.velocityY = velocity[1] / Gravitator.REFRATE
		self.affected = isAffected
		self.accelInX = 0
		self.accelInY = 0
		
		self.posX, self.posY = self.rect.center

	# gList is a list of other Gravitator sprites (or derived classes).
	# gameObjects is the corresponding group.
	def update(self, gameObjects):
		if self.affected:
			accelOutX = self.accelInX
			accelOutY = self.accelInY
			for g in gameObjects.sprites():
				if isinstance(g, Gravitator) and g != self:
					sqdist = ((g.posX - self.posX)*(g.posX - self.posX)+
						  (g.posY - self.posY)*(g.posY - self.posY))
					dist   = math.sqrt(sqdist)					
					accel  = g.mass * sqdist
					accelOutX += (accel * (g.posX - self.posX)) / dist
					accelOutY += (accel * (g.posY - self.posY)) / dist
			self.velocityX += accelOutX / Gravitator.REFRATE
			self.velocityY += accelOutY / Gravitator.REFRATE
		
		self.posX += self.velocityX
		self.posY += self.velocityY

		self.rect.center = (self.posX, self.posY)
		Animator.update(self)

	# adds an initial force to the model. The force can be reset by a call to addForce
	# with a force of 0.
	def addForce(self,f):
		self.accelInX = f[0] / self.mass
		self.accelInY = f[1] / self.mass

# This class represents the star that we have in the level.
# Pos is a 2-uple(X,Y)
class GStar(Gravitator):
	imageList = None

	def __init__(self, pos):
		if GStar.imageList == None:
			s0 = pygame.image.load(os.path.join("media","images","sun_00.png")).convert()
			s0.set_colorkey(0,RLEACCEL)
			s1 = pygame.image.load(os.path.join("media","images","sun_01.png")).convert()
			s1.set_colorkey(0,RLEACCEL)
			s2 = pygame.image.load(os.path.join("media","images","sun_02.png")).convert()
			s2.set_colorkey(0,RLEACCEL)
			GStar.imageList = [s0, s0, s0, s0, s1, s1, s1, s1, s2, s2, s2, s2, s1, s1, s1, s1]
		rect = GStar.imageList[0].get_rect()
		rect.center = pos

		Gravitator.__init__(self, GStar.imageList, rect, 0.0001, (0,0), False)

# A null-for-now class that has special properties when the
# object drawn leaves the area.
class AreaLeaver(Gravitator):
	pass

# A class for things that can die on collision.
# dielist is the animation for dying
# dieTypes is the types of objects on which CanDier dies when colliding.
# Renamed to the more generic "GameObject".
class GameObject(AreaLeaver):
	def __init__(self, imageList, dieList, pos, mass, velocity, isAffected=True,angle=0):
		AreaLeaver.__init__(self, imageList, pos, mass, velocity, isAffected, angle)
		self.dieList      = dieList
		self.alive        = True
		self.deathcounter = 0
		self.isKilled     = False

	def update(self, gameObjects):
		if self.alive:
			AreaLeaver.update(self, gameObjects)
		else:
			Animator.update(self)
			self.deathcounter += 1
			if self.deathcounter >= len(self.dieList):
				self.kill()

	# default is to show the dying animation. Can be overriden.
	# object is the object that collides.
	def hit(self, obj):
		if obj != self and self.alive:
			self.alive = False
			self.setSequence(self.dieList)
			if isinstance(self, Box):
                                mySound.playBXCrash()
                        elif isinstance(self, UFO):
                                mySound.playSPCrash()

	def kill(self):
		self.isKilled = True
		AreaLeaver.kill(self)
# The infamous BOX class. The array of tiles is loaded into memory, but they will rotate
# in different ways depending on the initialization...
# Box generates its own initial position and velocity.
# it possesses a flag that says if it has been caught by the UFO or not.
class Box(GameObject):
	imageList = None
	dieList = None
	def __init__(self, pos, velocity):
		if Box.imageList == None:
			b0 = pygame.image.load(os.path.join("media","images","box_00.png")).convert()
			b0.set_colorkey(0, RLEACCEL)
			b1 = pygame.image.load(os.path.join("media","images","box_01.png")).convert()
			b1.set_colorkey(0, RLEACCEL)
			b2 = pygame.image.load(os.path.join("media","images","box_02.png")).convert()
			b2.set_colorkey(0, RLEACCEL)
			b3 = pygame.image.load(os.path.join("media","images","box_03.png")).convert()
			b3.set_colorkey(0, RLEACCEL)
			Box.imageList = [b0, b0, b1, b1, b2, b2, b3, b3]
		if Box.dieList == None:
			d0 = pygame.image.load(os.path.join("media","images","box_die_00.png")).convert()
			d0.set_colorkey(0, RLEACCEL)
			d1 = pygame.image.load(os.path.join("media","images","box_die_01.png")).convert()
			d1.set_colorkey(0, RLEACCEL)
			d2 = pygame.image.load(os.path.join("media","images","box_die_02.png")).convert()
			d2.set_colorkey(0, RLEACCEL)
			Box.dieList = [d0, d0, d1, d1, d2, d2]
		if random.randint(1, 2) == 1:
			imgList = Box.imageList
		else:
			imgList = copy.copy(Box.imageList)
			imgList.reverse()

		rect = imgList[0].get_rect()
		rect.center = pos
		GameObject.__init__(self, imgList, Box.dieList, rect, 0.0000001, velocity, True)
                mySound.playBXApp()		

	# overload	
	def hit(self, obj):
		if isinstance(obj, (GStar, Box)):
			GameObject.hit(self, obj)

# And, last but not least, the almost as infamous class "UFO!"
class UFO(GameObject):
	UFOanimBase = None
	UFOdie      = None
	UFOgoFwd    = None
	UFOgoBwd    = None
	UFOshoot    = None

	def __init__(self, pos):
		if not UFO.UFOanimBase:
			u0 = pygame.image.load(os.path.join("media","images","UFO_nop.png")).convert()
			u0.set_colorkey(0)
			UFO.UFOanimBase = [u0]
		if not UFO.UFOdie:
			u0 = pygame.image.load(os.path.join("media","images","UFO_die_00.png")).convert()
			u0.set_colorkey(0)		
			u1 = pygame.image.load(os.path.join("media","images","UFO_die_01.png")).convert()
			u1.set_colorkey(0)		
			u2 = pygame.image.load(os.path.join("media","images","UFO_die_02.png")).convert()
			u2.set_colorkey(0)		
			u3 = pygame.image.load(os.path.join("media","images","UFO_die_03.png")).convert()
			u3.set_colorkey(0)		
			u4 = pygame.image.load(os.path.join("media","images","UFO_die_04.png")).convert()
			u4.set_colorkey(0)		
			u5 = pygame.image.load(os.path.join("media","images","UFO_die_05.png")).convert()
			u5.set_colorkey(0)
			UFO.UFOdie = [u0, u0, u0, u1, u1, u1, u2, u2, u2, u3, u3, u3, u4, u4, u4, u5, u5, u5]
		if not UFO.UFOgoFwd:
			u0 = pygame.image.load(os.path.join("media","images","UFO_fwd_00.png")).convert()
			u0.set_colorkey(0)		
			u1 = pygame.image.load(os.path.join("media","images","UFO_fwd_01.png")).convert()
			u1.set_colorkey(0)
			UFO.UFOgoFwd = [u0, u1]
		if not UFO.UFOgoBwd:
			u0 = pygame.image.load(os.path.join("media","images","UFO_rev_00.png")).convert()
			u0.set_colorkey(0)		
			u1 = pygame.image.load(os.path.join("media","images","UFO_rev_01.png")).convert()
			u1.set_colorkey(0)
			UFO.UFOgoBwd = [u0, u1]

		rect = UFO.UFOanimBase[0].get_rect()
		rect.center = pos
		self.myGrabber = None		
		self.anglespeed = 0
		self.angleacc   = 0
		self.thrust     = 0
		self.numBoxes   = 0
		GameObject.__init__(self, UFO.UFOanimBase, UFO.UFOdie, rect, 0.000001, (0,0), True, 0)

	# shiny update! it's an initial speed that is given by the cannon.
	# A shot has been done. THe force sent by the ControlPanel is added to the default force.
	def fire(self, force,gameObjects):
		self.numBoxes -= 1
		B = Box((self.posX + math.cos(math.radians(self.angle))*40, self.posY - math.sin(math.radians(self.angle))*40),
				(self.velocityX+200*(1+force)*math.cos(math.radians(self.angle)), self.velocityY - 200*(1+force)*math.sin(math.radians(self.angle))))
		gameObjects.add(B)
		mySound.playSPFire()

	def fireGrab(self,gameObjects):
		if not self.myGrabber:
			self.myGrabber = Grabber(self, self.rect.center, 200, self.angle - 180, 250, gameObjects)
			gameObjects.add(self.myGrabber)
			mySound.playGBOut()
		else:
			self.myGrabber.retract()

	def startRotLeft(self):
		self.angleacc = 0.5

	def startRotRight(self):
		self.angleacc = -0.5

	def stopRot(self):
		self.angleacc = 0

	# at last!
	# note sign inversion, due to screen stuff.
	def startThruFwd(self):
		self.setSequence(UFO.UFOgoFwd)
		self.addForce((math.cos(math.radians(self.angle)) / 100000, - math.sin(math.radians(self.angle)) / 100000))
                mySound.playSPMove()

	def startThruBwd(self):
		self.setSequence(UFO.UFOgoBwd)
		self.addForce((- math.cos(math.radians(self.angle)) / 100000, math.sin(math.radians(self.angle)) / 100000))
                mySound.playSPMove()

	def stopThru(self):
		self.setSequence(UFO.UFOanimBase)
		self.addForce((0,0))
                mySound.stopSPMove()

	def update(self, gameObjects):
		self.anglespeed += self.angleacc
		# angle is used by "animator"
		self.angle += self.anglespeed
		GameObject.update(self, gameObjects)
		# cheat!
		self.rect = self.image.get_rect()
		self.rect.center = (self.posX, self.posY)
	
	def hit(self, obj):
		if isinstance(obj, GStar):
			r = self.rect
			# really hit? Notice how small is the new rectangle, 22*22
			if r.inflate(-(r.width - 22), -(r.height-22)).colliderect(obj.rect):
				GameObject.hit(self, obj)
				if self.myGrabber:
					self.myGrabber.kill()
		elif isinstance(obj, Box):
			r = self.rect
			# eviiiil!
			if r.inflate(-(r.width - 22), -(r.height-22)).colliderect(obj.rect):
				# well here we destroy the box...
				GameObject.hit(obj, self)
				pass
		elif obj == self.myGrabber and obj.isRetracting:
			r = self.rect
			# eviiiil!
			if r.inflate(-(r.width - 22), -(r.height-22)).colliderect(obj.rect):
				if obj.gotOne:
					self.numBoxes += 1
				obj.kill()
				self.myGrabber = None
                                mySound.playGBIn()
			
			

# The Grabber. It is instantiated at runtime by the UFO.
# For now we won't care about the little shiny boxes.
# it is not affected by gravity (or is it?)
# it is defined by a position and a velocity and an UFO and an angle.
# here velocity is a single integer.
# DMax is the maximal distance between the grabber and the UFO.
class Grabber(GameObject):
	GrabberNormal = None
	GrabberGotcha = None

	def __init__(self, anUFO, pos, velocity, angle, DMax, gameObjects):
		if not Grabber.GrabberNormal:
			g = pygame.image.load(os.path.join("media","images","grabber_empty.png")).convert()
			g.set_colorkey(0)
			Grabber.GrabberNormal = [g]
		if not Grabber.GrabberGotcha:
			g = pygame.image.load(os.path.join("media","images","grabber_full.png")).convert()
			g.set_colorkey(0)
			Grabber.GrabberGotcha = [g]
		self.normAnim = [pygame.transform.rotate(Grabber.GrabberNormal[0], angle)]
		self.gotAnim  = [pygame.transform.rotate(Grabber.GrabberGotcha[0], angle)]

		rect = self.normAnim[0].get_rect()
		rect.center = pos
		self.anUFO = anUFO
		self.DMax  = DMax
		self.isRetracting = False
		self.velRetract = 10
		self.gotOne = False
		velX = math.cos(math.radians(angle))*velocity
		velY = - math.sin(math.radians(angle))*velocity
		# note: we "fake" an angle of 0 degrees since it has been already
		# rotated and keep its rotation until the very end.
		GameObject.__init__(self, self.normAnim, [], rect, 0.0000000001, (velX, velY), 0)

		# and finally we add the dots.
		self.chainL = []
		for i in range(2, 10):
			D = GrabberDot(anUFO, self, i)
			gameObjects.add(D)
			self.chainL.append(D)


	def update(self, gameObjects):
		dist = math.sqrt((self.posX - self.anUFO.posX)*(self.posX - self.anUFO.posX)+
			(self.posY - self.anUFO.posY)*(self.posY - self.anUFO.posY))
		if not self.isRetracting and (dist > self.DMax or self.gotOne):
			self.retract()
		if self.isRetracting:
			# cheat! we modify velocity directly.
			angle = math.acos((- self.posX + self.anUFO.posX) / dist)
			velX = abs(math.cos(angle) * self.velRetract)
			velY = abs(math.sin(angle) * self.velRetract)
			if self.posX > self.anUFO.posX:
				velX = - velX
			if self.posY > self.anUFO.posY:
				velY = - velY
			self.velocityX = velX
			self.velocityY = velY
		GameObject.update(self, gameObjects)

	def hit(self, obj):
		# we remove the box.
		if isinstance(obj, Box) and not self.gotOne:
			rect = self.rect.inflate(- (self.rect.width - 10), - (self.rect.height - 10))
			if rect.colliderect(obj.rect):
				self.setSequence(self.gotAnim)
				obj.kill()
				self.gotOne = True
                                mySound.playGBGrab()

	def retract(self):
                if not self.isRetracting:
        		self.isRetracting = True
                        mySound.playGBMax()
	# we also remove the dots here.
	def kill(self):
		for o in self.chainL:
			o.kill()
		GameObject.kill(self)

# grabber dot. Just some nifty visuals.
# pos is an integer from 0 to 10.
class GrabberDot(pygame.sprite.Sprite):
	GrabberImg = None
	def __init__(self, anUFO, aGrabber, pos):
		pygame.sprite.Sprite.__init__(self)
		if not GrabberDot.GrabberImg:
			GrabberDot.GrabberImg = pygame.image.load(os.path.join("media","images", "grabber_chain.png")).convert()
			GrabberDot.GrabberImg.set_colorkey(0)
		self.image = GrabberDot.GrabberImg
		self.rect = self.image.get_rect()
		self.pos = pos
		self.aGrabber = aGrabber
		self.anUFO = anUFO
		posX = (aGrabber.posX + (anUFO.posX - aGrabber.posX) * pos / 11)
		posY = (aGrabber.posY + (anUFO.posY - aGrabber.posY) * pos / 11)
		self.rect.center = (posX, posY)

	# we don't really care about foo...
	def update(self, foo):
		posX = (self.aGrabber.posX + (self.anUFO.posX - self.aGrabber.posX) * self.pos / 11)
		posY = (self.aGrabber.posY + (self.anUFO.posY - self.aGrabber.posY) * self.pos / 11)
		self.rect.center = (posX, posY)
		pygame.sprite.Sprite.update(self)

# The control bar (left on the screen).
# takes an UFO as parameter (for now...).
class ControlBar(pygame.sprite.Sprite):
	CBImg = None
	PBC   = None
	PBN   = None
	BoxI  = None
	def __init__(self, anUFO):
		pygame.sprite.Sprite.__init__(self)
		if not ControlBar.CBImg:
			ControlBar.CBImg = pygame.image.load(os.path.join("media","images", "lpanel.png")).convert()
		if not ControlBar.PBC:
			p0 = pygame.image.load(os.path.join("media","images", "powerbar_low.png")).convert()
			p0.set_colorkey(0)
			p1 = pygame.image.load(os.path.join("media","images", "powerbar_med.png")).convert()
			p1.set_colorkey(0)
			p2 = pygame.image.load(os.path.join("media","images", "powerbar_hig.png")).convert()
			p2.set_colorkey(0)
			ControlBar.PBC=(p0,p1,p2)			
		if not ControlBar.PBN:
			ControlBar.PBN = pygame.image.load(os.path.join("media","images", "powerbar_off.png")).convert()
			ControlBar.PBN.set_colorkey(0)
		if not ControlBar.BoxI:
			ControlBar.BoxI = pygame.image.load(os.path.join("media","images","box_00.png")).convert()
			ControlBar.BoxI.set_colorkey(0)
		self.image = ControlBar.CBImg
		self.rect = self.image.get_rect()
		self.rect.topleft = (0,0)
		self.anUFO = anUFO
		self.power = 0
		self.maxPower = 47
		self.charging = False

	def startShooting(self):
		if self.anUFO.numBoxes >= 1:
			self.charging = True
                        mySound.playCNLoad()

	def stopShooting(self, gameObjects):
		if self.charging:
			self.charging = False
			# we send a "power" value between 0 and 1... Normalised, in a sense.
			mySound.stopCNLoad()
			self.anUFO.fire(self.power / self.maxPower, gameObjects)
			self.power = 0

	def update(self, gameObjects):
		if self.power >= self.maxPower:
			self.stopShooting(gameObjects)
		for i in range(0, self.power):
			if i < 16:
				self.image.blit(ControlBar.PBC[0], (4, 235 - i * 5))
			elif i < 32:
				self.image.blit(ControlBar.PBC[1], (4, 235 - i * 5))
			else:
				self.image.blit(ControlBar.PBC[2], (4, 235 - i * 5))
		for i in range(self.power, self.maxPower):
			self.image.blit(ControlBar.PBN, (4, 235 - i*5))
		m = self.anUFO.numBoxes
		if m > 15:
			m = 15
		for i in range(0, m):
			self.image.blit(ControlBar.BoxI, (2, 460 - i * 15))
		for i in range(m, 15):
			self.image.fill(0, pygame.Rect(2, 460 - i * 15, 15, 15))
		if self.charging:
			self.power += 1
		pygame.sprite.Sprite.update(self)

# Last useful class for the game: The portal!!!
class Portal(GameObject):
	UpDoor = None
	DoDoor = None
	def __init__(self, pos, height):
		if not Portal.UpDoor:
			Portal.UpDoor = pygame.image.load(os.path.join("media","images", "portal_up.png")).convert()
			Portal.UpDoor.set_colorkey(0, RLEACCEL)
		if not Portal.DoDoor:
			Portal.DoDoor = pygame.image.load(os.path.join("media","images", "portal_dn.png")).convert()
			Portal.DoDoor.set_colorkey(0, RLEACCEL)
		sl = []
		for i in range (20, height, 5):
			s = pygame.Surface((10, i))
			s.set_colorkey(0)
			s.blit(Portal.UpDoor, (0,0))
			s.blit(Portal.DoDoor, (0, i - 11))
			sl.append(s)
		sl2 = copy.copy(sl)
		sl2.reverse()
		sl.extend(sl2)
		self.imgList = sl
		rect = sl[0].get_rect()
		rect.center = pos
		self.pos = pos
		self.score = 0
		GameObject.__init__(self, self.imgList, [], rect, 0, (0,0), False)

	def update(self, gameObjects):
		self.rect = self.image.get_rect()
		self.rect.center = self.pos
		GameObject.update(self, gameObjects)


	def hit(self, obj):
		# we've been hit by a box!
		if isinstance(obj, Box):
			self.score += 1
			self.setSequence(self.imgList)
			obj.kill()
			mySound.playBXThru()

# The scoreboard. It is initialized with the time, the clock handling the game,
# the portal and a portal. It will show the score on the left and the time left on the
# right. One method returns true when time is over. The other returns the final score.
class ScoreBoard(pygame.sprite.Sprite):
	ScoreImage = None
	def __init__(self, duration, aPortal, aClock):
		pygame.sprite.Sprite.__init__(self)
		if not ScoreBoard.ScoreImage:
			ScoreBoard.ScoreImage = pygame.image.load(os.path.join("media","images", "scoreboard.png")).convert()
			ScoreBoard.ScoreImage.set_colorkey(0, RLEACCEL)
		self.oldTs=None
		self.oldTd=None
		self.myPortal = aPortal
		self.myClock  = aClock
		self.duration = (duration + 1)* 1000 # millisecs
		self.image = ScoreBoard.ScoreImage
		self.rect  = self.image.get_rect()
		self.myScore = aPortal.score - 1
		self.rect.bottomright = (DX - 1, DY - 1)
		self.myFont = pygame.font.Font("afont.ttf", 20)
		self.printOn(aPortal.score, duration * 1000)


	def update(self, foo):
		score    = self.myPortal.score
		duration = self.duration - self.myClock.get_time()
		self.printOn(score, duration)
		self.duration = duration
		pygame.sprite.Sprite.update(self)
		
	def printOn(self, score, duration):
		if score != self.myScore:
			ts = self.myFont.render("%03d"%(score), True, [255,255,255], [3,3,3])
			self.myScore = score
			if self.oldTs:
				self.oldTs.fill([3,3,3])
				self.image.blit(self.oldTs, (5,5))
			self.oldTs = ts
			self.image.blit(ts, (5,5))
		if duration / 100 != self.duration / 100:
			if duration < 10000:
				td = self.myFont.render("%02.1f"%(duration/1000.0),True, [255,0,0],[3,3,3])
			else:
				td = self.myFont.render("%02.1f"%(duration/1000.0),True, [255,255,255],[3,3,3])
			self.duration = duration
			if self.oldTd:
				self.oldTd.fill([3,3,3])
				self.image.blit(self.oldTd,(40,5))
			self.oldTd = td
			self.image.blit(td, (40, 5))

	def isOver(self):
		return (self.duration <= 0.001)

	def score(self):
		return self.myScore

	def timeLeft(self):
		return max(self.duration/1000, 0)

# DEPRECATED DOC
# THe main "level" class, that generates a level in which the player can, well, play.
# duration in seconds, posUFO is (X,y), posstars is a list of (X, y), probbox is an integer
# between 0 (a otn of boxes) and 1000 (no box).
# heightPortal is a height in Pixels.
class Level:
	def __init__(self, keyOpts, levelOpts):
		global DX, DY, screen, background,mySound	

		DX=myglobals.DX
		DY=myglobals.DY
		screen=myglobals.screen
		background=myglobals.background
		mySound=myglobals.mySound

		self.duration = levelOpts["lenplay"]
		self.gameObjects = pygame.sprite.RenderUpdates()

		self.thePortal = Portal((634, 240), levelOpts["portalsize"])
		self.gameObjects.add(self.thePortal)
		
		if levelOpts["stars"]==1:
			self.gameObjects.add(GStar((320, 240)))
			self.theUFO = UFO((30, 240))
		elif levelOpts["stars"]==2:
			self.gameObjects.add(GStar((200, 200)))
			self.gameObjects.add(GStar((440, 280)))
			self.theUFO = UFO((320, 240))
		elif levelOpts["stars"]==3:
			self.gameObjects.add(GStar((214, 100)))
			self.gameObjects.add(GStar((214, 380)))
			self.gameObjects.add(GStar((428, 240)))
			self.theUFO = UFO((620, 240))
			 			
		self.theCtl = ControlBar(self.theUFO)
		self.gameObjects.add(self.theUFO)
		self.gameObjects.add(self.theCtl)

		self.probBox = levelOpts["boxprob"]
		self.keys = keyOpts

	# Useful if keys have been remapped during the level's execution.
	def getKeys(self):
		return self.keys

	# returns (time spent, score).		
	def run(self):
		C = pygame.time.Clock()
		C.tick()
		C.tick()
		theScoreBoard = ScoreBoard(self.duration, self.thePortal, C)
		self.gameObjects.add(theScoreBoard)
		theUFO = self.theUFO
		theCtl = self.theCtl
		gameObjects = self.gameObjects
		leaveThisShit = False
		keys = self.keys
		screen.fill(0)
		pygame.display.flip()
		mySound.playTune("game.wav")
		while not (theUFO.isKilled or theScoreBoard.isOver() or leaveThisShit):
			# Clearing and updating.
			gameObjects.update(gameObjects)

			# Collisions management.
			# Note the lambda function syntax.
			for sp in gameObjects.sprites():
				if isinstance(sp, GameObject):
					# faster with a map, isn't it?
					map(sp.hit, pygame.sprite.spritecollide(sp, gameObjects, False))
	
			# Player action management.
			for e in pygame.event.get():
				if e.type == KEYDOWN:
					if e.key == K_ESCAPE:
						M = MainMenu(screen, True, self.keys, None)
						M.show()
						self.keys = M.getAll()[0]
						keys = self.keys
						if M.hasQuit():
							leaveThisShit=True
						C.tick()
					elif e.key == keys["left"]:
						theUFO.startRotLeft()
					elif e.key == keys["right"]:
						theUFO.startRotRight()
					elif e.key == keys["up"]:
						theUFO.startThruFwd()
					elif e.key == keys["down"]:
						theUFO.startThruBwd()
					elif e.key == keys["release"]:
						theUFO.fireGrab(gameObjects)
					elif e.key == keys["shoot"]:
						theCtl.startShooting()
				elif e.type == KEYUP:
					if e.key in [keys["left"], keys["right"]]:
						theUFO.stopRot()
					elif e.key in [keys["up"], keys["down"]]:
						theUFO.stopThru()
                                        elif e.key == keys["shoot"]:
						theCtl.stopShooting(gameObjects)
	
			pygame.event.pump()
						

			# Random box appears!
			if random.randint(0, 1000) <= self.probBox:
				b = Box((random.randint(0, DX), random.randint(0, DY)), (random.randint(20,100), random.randint(20,100)))
				gameObjects.add(b)
	
			# Drawing the stuff and waiting...
			rectlist = gameObjects.draw(screen)
			pygame.display.update(rectlist)
			C.tick(30)
			gameObjects.clear(screen, background)

		self.gameObjects.empty()
		return (theScoreBoard.score(),self.duration - theScoreBoard.timeLeft())
